from dataclasses import dataclass
from typing import Type, Tuple

import pandas as pd

from feature_forge.domain.entities.entity_model import EntityModel
from feature_forge.domain.entities.feature import Feature
from feature_forge.domain.entity_models.transaction import Transaction
from feature_forge.domain.features.transaction_amount_change import \
    TransactionAmountChange


@dataclass(frozen=True)
class TransactionRoundedAmountChange(Feature):
    col_name: str = "transaction_rounded_amount_change"
    description: str = "rounded difference between withdrawal amount and desposit " \
                  "amount"
    entity_model: Type[EntityModel] = Transaction
    dependencies: Tuple[Feature] = (TransactionAmountChange(),)

    def row_compute_fn(self, row: pd.Series):
        row[self.col_name] = row[TransactionAmountChange.col_name] / 1000
        return row

    def table_compute_fn(self, data: pd.DataFrame):
        return self.row_compute_fn(data)




from dataclasses import dataclass
from typing import Type, List, Callable

import pandas as pd
from pydot import Dot

from feature_forge.domain.entities.dataset import Dataset
from feature_forge.domain.entities.entity_model import EntityModel


@dataclass
class Pipeline:
    dataset: Dataset
    entity: Type[EntityModel]
    flow: Callable[[pd.DataFrame], pd.DataFrame]
    graph: Dot
    requested_features: List[str]
    all_features: List[str]


from .halved_rounded_amount_change import *
from .rounded_amount_change import *
from .rounded_amount_change_difference import *
from .transaction_rounded_amount_change_diff_doubled import *
from .transaction_rounded_amount_change_diff_squared import *
from .transaction_month import TransactionMonth
from .transaction_amount_change import *

__all__ = [
    'TransactionHalfRoundedAmountChange',
    'TransactionMonth',
    'TransactionRoundedAmountChangeDiff',
    'TransactionAmountChange',
    'TransactionHalfRoundedAmountChange',
    'TransactionRoundedAmountChangeDiffDoubled',
    'TransactionRoundedAmountChangeDiffSquared',
]



from dataclasses import dataclass
from typing import Type, Protocol, Tuple, runtime_checkable, ClassVar

import pandas as pd

from feature_forge.domain.entities.entity_model import EntityModel


@dataclass(frozen=True)
class Dataset:
    supported_entity_models: ClassVar[Tuple[Type[EntityModel]]]
    data: pd.DataFrame

    def map_rows_to_entity(self, entity: Type[EntityModel]):
        """Maps the dataframe such that rows to entities is 1-to-1."""
        ...



from functools import wraps
from typing import Union

import pandas as pd

from feature_forge.domain.entities.feature import Feature


class FeatureValidator:

    def add_runtime_validation(self, compute_fn: Union[Feature.row_compute_fn,
                             Feature.table_compute_fn]):
        return self._validate_feature_compute_fn(compute_fn)

    def _validate_feature_compute_fn(self, func):
        inst = func.__self__

        @wraps(func)
        def wrapper(data: Union[pd.DataFrame, pd.Series]) -> Union[
            pd.DataFrame, pd.Series]:
            if inst.col_name in data:
                raise KeyError(
                    f"Feature {inst.col_name} already exists in the "
                    f"dataset. Rename this column or the feature col_name")
            result = func(data)
            if not inst.col_name in result:
                raise KeyError(f"Feature {inst.col_name} has not created "
                               f"the feature in the dataset.")
            return result

        return wrapper



import itertools
from typing import List, Type, Dict

import networkx as nx
from networkx import DiGraph
from pydot import Dot

from feature_forge.domain.entities.entity_model import EntityModel
from feature_forge.domain.entities.feature import Feature


class DAGBuilder:

    def build_simple_graph(self,
                           features: List[Feature],
                           entity: Type[EntityModel]):
        graph = nx.DiGraph()
        for feature in features:
            graph.add_edges_from([(entity, feature)])
            graph = self._add_dependencies(graph, entity, feature)

        nx.is_directed_acyclic_graph(graph)
        graph = nx.transitive_reduction(graph)
        return graph

    def _follow_path_until_split(self, graph, node, inline_edges=None):
        inline_edges = inline_edges or []

        out_edges = graph.out_edges(node)
        in_edges = graph.in_edges(node)

        number_of_out_edges = len(out_edges)
        number_of_in_edges = len(in_edges)

        if number_of_in_edges < 2:
            inline_edges.append(node)
        else:
            return inline_edges

        if number_of_out_edges != 1:
            return inline_edges

        out_node = list(out_edges)[0][1]
        return self._follow_path_until_split(graph, out_node, inline_edges=inline_edges)

    def build_grouped_graph(self,
                            single_path_graph: DiGraph):
        graph = single_path_graph

        inline_nodes = []
        for node in graph.nodes:
            sequenced_nodes = self._follow_path_until_split(graph, node)
            if len(sequenced_nodes) > 1:
                inline_nodes.append(tuple(sequenced_nodes))

        inline_nodes = sorted(inline_nodes, key=lambda x: -len(x))

        supernodes = []
        for group in inline_nodes:
            group_len = len(group)
            group_is_sublist = False
            for supernode in supernodes:
                supernode_len = len(supernode)
                if supernode == group:
                    group_is_sublist = True
                for i in range(supernode_len - group_len + 1):
                    if supernode[i:i+group_len] == group:
                        group_is_sublist = True
            if not group_is_sublist:
                supernodes.append(group)

        for supernode in supernodes:
            out_nodes = [e[1] for e in graph.out_edges(supernode[-1])]
            in_nodes = [e[0] for e in graph.in_edges(supernode[0])]

            for node in supernode:
                graph.remove_node(node)

            graph.add_node(supernode)
            graph.add_edges_from([(i, supernode) for i in in_nodes])
            graph.add_edges_from([(supernode, o) for o in out_nodes])

        return nx.transitive_reduction(graph)

    def build_pydot(self,
                    graph: DiGraph,
                    label_map: Dict[str, str]) -> Dot:
        dot_graph = nx.nx_pydot.to_pydot(graph)
        for node in dot_graph.get_nodes():
            node.set_label(label_map[node.get_name().replace('"', '')])
        return dot_graph

    def _add_dependencies(self, graph, entity, feature, dependencies=None):
        dependencies = dependencies or []
        for dep in feature.dependencies:
            graph.add_edges_from([(entity, dep), (dep, feature)])
            self._add_dependencies(graph, entity, dep,
                                   dependencies=dependencies)
        return graph





from dataclasses import dataclass
from typing import Type, Tuple

import numpy as np
import pandas as pd

from feature_forge.domain.entities.entity_model import EntityModel
from feature_forge.domain.entities.feature import Feature
from feature_forge.domain.entity_models.transaction import \
    Transaction


@dataclass(frozen=True)
class TransactionAmountChange(Feature):
    col_name: str = "transaction_amount_change"
    description: str = "difference between witdrawal amount and deposit amount"
    entity_model: Type[EntityModel] = Transaction
    dependencies: Tuple[Feature] = tuple()

    def row_compute_fn(self, row: pd.Series):
        original_withdrawal_amt = row[Transaction['withdrawal_amount']]
        original_deposit_amount = row[Transaction['deposit_amount']]
        withdrawal_amount = 0 if np.isnan(original_withdrawal_amt) \
            else original_withdrawal_amt
        deposit_amount = 0 if np.isnan(original_deposit_amount) \
            else original_deposit_amount
        row[self.col_name] = withdrawal_amount - deposit_amount
        return row

    def table_compute_fn(self, data: pd.DataFrame) -> pd.Series:
        data[self.col_name] = data[Transaction['withdrawal_amount']].fillna(0) - data[Transaction['deposit_amount']].fillna(0)
        return data


import pandas as pd

from feature_forge.domain.entities.pipeline import Pipeline


class PipelineRunner:

    def run(self, pipeline: Pipeline, copy: bool) -> pd.DataFrame:
        entity_rows_df = pipeline.dataset.map_rows_to_entity(
            pipeline.entity)
        input_data = entity_rows_df.copy() if copy else entity_rows_df
        result = pipeline.flow(input_data)
        return result



from typing import Type

import pandera

from feature_forge.domain.entities.dataset import Dataset


class DatasetModelBuilder:

    def build(self, dataset: Dataset) -> Type[pandera.DataFrameModel]:

        class DatasetModel(pandera.DataFrameModel):
            class Config:
                dtype = dataset.entity_model
                coerce = True

        return DatasetModel



from dataclasses import dataclass
from typing import Type, ClassVar

from feature_forge.domain.entities.dataset import Dataset
from feature_forge.domain.entities.entity_model import EntityModel
from feature_forge.domain.entity_models.account import Account
from feature_forge.domain.entity_models.transaction import \
    Transaction


@dataclass(frozen=True)
class BankDataset(Dataset):
    supported_entity_models = (Transaction, Account)

    def map_rows_to_entity(self, entity: Type[EntityModel]):
        if entity == Transaction:
            return self.data
        if entity == Account:
            return self.data.groupby(Account['account_no'])




from typing import Type, Tuple, Optional

import pandas as pd
from pydantic import Field
from pydantic.dataclasses import dataclass

from feature_forge.domain.entities.entity_model import EntityModel
from feature_forge.domain.entities.feature import Feature
from feature_forge.domain.entity_models.transaction import \
    Transaction


@dataclass(frozen=True)
class TransactionMonth(Feature):
    col_name: str = "value_month"
    description: str = "month of the value date for the transaction"
    entity_model: Type[EntityModel] = Transaction
    dependencies: Optional[Tuple[Feature]] = Field(default_factory=tuple)

    def row_compute_fn(self, data: pd.Series):
        data[self.col_name] = data[Transaction['value_date']].month
        return data

    def table_compute_fn(self, data: pd.DataFrame):
        data[self.col_name] = data[Transaction['value_date']].dt.month
        return data



from dependency_injector import containers, providers

from feature_forge.domain.dag_builder import DAGBuilder
from feature_forge.domain.dataset_model_builder import DatasetModelBuilder
from feature_forge.domain.feature_validator import FeatureValidator
from feature_forge.domain.pipeline_builder import PipelineBuilder
from feature_forge.domain.pipeline_runner import PipelineRunner


class DomainContainer(containers.DeclarativeContainer):

    pipeline_builder = providers.Factory(
        PipelineBuilder,
        providers.Factory(DatasetModelBuilder),
        providers.Factory(DAGBuilder),
        providers.Factory(FeatureValidator)
    )

    pipeline_runner = providers.Factory(
        PipelineRunner
    )


from datetime import datetime

from pydantic import Field

from feature_forge.domain.entities.entity_model import EntityModel


class Transaction(EntityModel):
    account_no: str = Field(alias="Account No")
    date: datetime = Field(alias="DATE")
    trn_details: str = Field(alias="TRANSACTION DETAILS")
    chq_no: float = Field(alias="CHQ.NO.")
    value_date: datetime = Field(alias="VALUE DATE")
    withdrawal_amount: float = Field(alias="WITHDRAWAL AMT")
    deposit_amount: float = Field(alias="DEPOSIT AMT")
    balance_amount: float = Field(alias="BALANCE AMT")


import io
from typing import Type, List, Callable

import matplotlib.pyplot
import pandas as pd
from PIL import Image
from networkx import DiGraph
from prefect import Flow
from pydantic import BaseModel
from pydot import Dot

from feature_forge.domain.entities.dataset import Dataset
from feature_forge.domain.entities.entity_model import EntityModel


class Recipe(BaseModel):
    dataset: Dataset
    entity: Type[EntityModel]
    flow: Callable[[pd.DataFrame], pd.DataFrame]
    graph: Dot
    requested_features: List[str]
    all_features: List[str]

    class Config:
        arbitrary_types_allowed = True
        frozen = True

    def __str__(self):
        return self.__repr__()

    def __repr_args__(self):
        return [(a, v) if a not in ['dataset']
                else (a, self.dataset.__class__.__name__)
                for a, v in super().__repr_args__()]

    def show(self):
        image = Image.open(io.BytesIO(self.graph.create_png()))
        image.show()



from dependency_injector import containers, providers

from feature_forge.domain.container import DomainContainer


class Container(containers.DeclarativeContainer):

    domain = providers.Container(
        DomainContainer
    )



from typing import List

from pydantic import BaseModel
from pydantic.fields import ModelField


class EntityModel(BaseModel):

    @classmethod
    def __class_getitem__(cls, item) -> str:
        field: ModelField = cls.__fields__.get(item)
        return field.alias

    @classmethod
    def columns(cls) -> List[str]:
        return [f.alias for f in cls.__fields__.values()]

    @classmethod
    @property
    def name(cls) -> str:
        return cls.__name__


from typing import Type, Tuple, runtime_checkable, Protocol, Optional, Callable

import pandas as pd
from pydantic.dataclasses import dataclass

from feature_forge.domain.entities.entity_model import EntityModel


class FeatureMeta(type):
    def __repr__(cls):
        return f'{cls.__name__}'


@dataclass(frozen=True)
class Feature:
    __metaclass__ = FeatureMeta
    col_name: str
    description: str
    entity_model: Type[EntityModel]
    dependencies: Tuple["Feature"]

    feature_name: Optional[str] = None
    row_level_function: Optional[Callable[[pd.Series], pd.Series]] = None
    table_level_function: Optional[Callable[[pd.DataFrame], pd.DataFrame]] = None

    @property
    def name(self) -> str:
        return self.feature_name if self.feature_name else self.__class__.__name__

    def row_compute_fn(self, row: pd.Series) -> pd.Series:
        if self.row_level_function:
            return self.row_level_function(row)
        raise NotImplementedError()

    def table_compute_fn(self, data: pd.DataFrame) -> pd.DataFrame:
        if self.table_level_function:
            return self.table_level_function(data)
        raise NotImplementedError()



from dataclasses import dataclass
from typing import Type, Tuple

import pandas as pd

from feature_forge.domain.entities.entity_model import EntityModel
from feature_forge.domain.entities.feature import Feature
from feature_forge.domain.entity_models.transaction import Transaction
from feature_forge.domain.features.rounded_amount_change_difference import \
    TransactionRoundedAmountChangeDiff


@dataclass(frozen=True)
class TransactionRoundedAmountChangeDiffSquared(Feature):
    col_name: str = "transaction_rounded_amount_change_diff_squared"
    description: str = "difference of the rounded difference between the withdrawal" \
                  " amount and the deposit amount squared"
    entity_model: Type[EntityModel] = Transaction
    dependencies: Tuple[Feature] = (TransactionRoundedAmountChangeDiff(),)

    def row_compute_fn(self, row: pd.Series):
        row[self.col_name] = row[TransactionRoundedAmountChangeDiff.col_name] ** 2
        return row



from functools import cached_property
from typing import List, Type

import pandas as pd

from feature_forge.application.container import Container
from feature_forge.domain.entities.dataset import Dataset
from feature_forge.domain.entities.feature import Feature
from feature_forge.domain.entities.pipeline import Pipeline
from feature_forge.domain.pipeline_builder import PipelineBuilder
from feature_forge.domain.pipeline_runner import PipelineRunner
from feature_forge.application.models.recipe import Recipe


class Blacksmith:

    def __init__(self):
        self._container = Container()
        self._container.wire([__name__])
        self._container.init_resources()

    @cached_property
    def _pipeline_builder(self) -> PipelineBuilder:
        return self._container.domain.pipeline_builder()

    @cached_property
    def _pipeline_runner(self) -> PipelineRunner:
        return self._container.domain.pipeline_runner()

    def build_recipes(self,
                      dataset: Dataset,
                      features: List[Feature],
                      concurrent_paths: bool = False) -> List[Recipe]:
        pipelines = self._pipeline_builder.build(
            dataset=dataset,
            features=features,
            concurrent_paths=concurrent_paths
        )
        return [Recipe(dataset=pipeline.dataset,
                       entity=pipeline.entity,
                       flow=pipeline.flow,
                       graph=pipeline.graph,
                       requested_features=pipeline.requested_features,
                       all_features=pipeline.all_features,
                       ) for pipeline in pipelines]

    def forge(self, recipe: Recipe, copy: bool = True) -> pd.DataFrame:
        return self._pipeline_runner.run(pipeline=Pipeline(
            dataset=recipe.dataset,
            entity=recipe.entity,
            flow=recipe.flow,
            graph=recipe.graph,
            requested_features=recipe.requested_features,
            all_features=recipe.all_features
        ), copy=copy)






import itertools
from copy import deepcopy
from itertools import groupby
from typing import List, Tuple, Union

import networkx as nx
import pandas as pd
from prefect import flow, task

from feature_forge.domain.composite_task import CompositeTask
from feature_forge.domain.dag_builder import DAGBuilder
from feature_forge.domain.dataset_model_builder import DatasetModelBuilder
from feature_forge.domain.entities.dataset import Dataset
from feature_forge.domain.entities.feature import Feature
from feature_forge.domain.entities.pipeline import Pipeline
from feature_forge.domain.feature_validator import FeatureValidator


class PipelineBuilder:

    def __init__(self,
                 dataset_model_builder: DatasetModelBuilder,
                 dag_builder: DAGBuilder,
                 feature_validator: FeatureValidator):
        self._dataset_model_builder = dataset_model_builder
        self._dag_builder = dag_builder
        self._feature_validator = feature_validator

    def build(self,
              dataset: Dataset,
              features: List[Feature],
              concurrent_paths: bool) -> List[Pipeline]:

        entity_feature_map = {entity: list(features)
                              for entity, feature_group
                              in groupby(features,
                                         lambda f: f.entity_model)}
        unsupported_entities = set(entity_feature_map.keys()).difference(
            dataset.supported_entity_models)
        if unsupported_entities:
            raise AttributeError('Features depend on entities not '
                                 f'supported by this dataset ({unsupported_entities = })')
        pipelines = []
        for entity in entity_feature_map.keys():
            if concurrent_paths:
                pipeline = self._build_concurrent_paths_flow(
                    dataset=dataset,
                    entity=entity,
                    entity_feature_map=entity_feature_map
                )
                pipelines.append(pipeline)
            else:
                pipeline = self._build_single_path_flow(
                    dataset=dataset,
                    entity=entity,
                    entity_feature_map=entity_feature_map
                )
                pipelines.append(pipeline)
        return pipelines

    def _build_single_path_flow(self, dataset, entity, entity_feature_map):
        graph = self._dag_builder.build_simple_graph(
            features=entity_feature_map[entity],
            entity=entity
        )
        ordered_features = [f for f in nx.topological_sort(graph) if
                            isinstance(f, Feature)]

        func_map = self._build_feature_func_map(ordered_features)
        tasks = self._create_composite_tasks(func_map)

        return Pipeline(
            dataset=dataset,
            entity=entity,
            flow=self._build_pipeline(
                ordered_tasks=tasks
            ),
            graph=self._dag_builder.build_pydot(
                graph=graph,
                label_map={str(n): n.name
                           for n in
                           nx.nodes(graph)}),
            requested_features=[f.col_name for f in
                                entity_feature_map[entity]],
            all_features=[f.col_name for f in ordered_features]

        )

    def _create_composite_tasks(self, func_map):
        tasks = []
        for key, grouped in groupby(func_map, key=lambda x: x[0]):
            fs = [g[1] for g in grouped]
            if key == 'row':
                tasks.append(CompositeTask(
                    type=key, funcs=[
                        self._feature_validator.add_runtime_validation(
                            f.row_compute_fn) for f in fs],
                    features=[f.col_name for f in fs]))
            else:
                tasks += [
                    CompositeTask(
                        type=key,
                        funcs=[
                            self._feature_validator.add_runtime_validation(
                                f.table_compute_fn)
                        ], features=[f.col_name]
                    )
                    for f in fs
                ]
        return tasks

    def _build_feature_func_map(self, ordered_features):
        func_map = []
        for feature in ordered_features:
            if not isinstance(feature, Feature):
                continue
            if feature.table_compute_fn.__func__ != \
                    Feature.table_compute_fn:
                func_map.append(('table', feature))
            else:
                func_map.append(('row', feature))
        return func_map

    def _build_concurrent_paths_flow(self, dataset, entity, entity_feature_map
                                     ):
        flows = []
        graph = self._dag_builder.build_simple_graph(
            features=entity_feature_map[entity],
            entity=entity)
        group_graph = self._dag_builder.build_grouped_graph(deepcopy(graph))
        all_features = []
        for group in nx.topological_sort(group_graph):
            predecessors = list(group_graph.in_edges(group))
            predecessor_nodes = tuple(p[0] for p in predecessors)
            group = (group,) if not isinstance(group, tuple) else group
            func_map = self._build_feature_func_map(group)
            grouped_ordered = self._create_composite_tasks(func_map)

            if grouped_ordered:
                flows.append((predecessor_nodes,
                              self._build_group_pipeline(group,
                                                         grouped_ordered)))
                all_features.append([t.features for t in grouped_ordered])

        pipeline = Pipeline(
            dataset=dataset,
            entity=entity,
            flow=self._build_main_pipeline(flows, entity),
            graph=self._dag_builder.build_pydot(
                graph=graph,
                label_map={
                    str(node): self._get_name_from_group_node(
                        node)
                    for node in nx.nodes(graph)}
            ),
            requested_features=[f.col_name for f in
                                entity_feature_map[entity]],
            all_features=[f for f in
                          self._flatten_feature_list(all_features)]
        )
        return pipeline

    def _flatten_feature_list(self, lst):
        return [item for sublist in lst for item in
                (self._flatten_feature_list(sublist) if isinstance(sublist,
                                                                   list) else [
                    sublist])]

    def _build_main_pipeline(self, flows, entity):

        def pipeline(x):
            dfs = {entity.name: x}
            original_cols = x.columns
            for predecessor_nodes, flow in flows:
                kwargs = {}
                wait_for = [self._get_name_from_group_node(p)
                            for p in predecessor_nodes] or None
                if wait_for:
                    kwargs['wait_for'] = [dfs[n] for n in wait_for]

                dfs[flow.name] = flow(x, **kwargs)
            dfs[entity.name] = dfs[entity.name].loc[:, original_cols]
            return self._concat_task(dfs=dfs.values(),
                                     wait_for=[dfs[k] for k in dfs.keys()])

        return flow(name=f"forging-{entity.name}-features-concurrent")(
            pipeline)

    @staticmethod
    @task(name='combine')
    def _concat_task(dfs):
        return pd.concat(dfs, axis=1)

    def _get_name_from_group_node(self,
                                  group: Union[Feature, Tuple[Feature]]):
        if not isinstance(group, tuple):
            return group.name
        return ' --} '.join(f.name for f in group)

    def _build_group_pipeline(self,
                              node: Tuple[Feature],
                              ordered_tasks: List[CompositeTask]):
        def pipeline(x):
            for t in ordered_tasks:
                x = t.prefect_task(x)
            return x[
                list(itertools.chain(*[t.features for t in ordered_tasks]))]

        return flow(name=self._get_name_from_group_node(node))(pipeline)

    def _build_pipeline(self, ordered_tasks: List[CompositeTask]):

        def pipeline(x):
            for t in ordered_tasks:
                x = t.task(x)
            return x

        return pipeline

    def _build_prefect_pipeline(self,
                                name: str,
                                ordered_tasks: List[CompositeTask]):

        def pipeline(x):
            for t in ordered_tasks:
                x = t.prefect_task(x)
            return x

        return flow(name=name)(pipeline)



from dataclasses import dataclass
from functools import reduce, cached_property
from typing import List, Callable, Union

import pandas as pd
from prefect import task

from feature_forge.domain.entities.feature import Feature


@dataclass
class CompositeTask:
    type: str
    funcs: List[Union[Feature.row_compute_fn, Feature.table_compute_fn]]
    features: List[str]

    def _build_elementwise_function(self):
        def elementwise_fn(x):
            x = x.apply(self.fn, axis=1)
            return x

        return elementwise_fn

    def _build_vectorised_function(self):
        def vectorized_fn(x):
            x = self.fn(x)
            return x

        return vectorized_fn

    @cached_property
    def prefect_task(self):
        return task(name=self.name)(self.task)

    @cached_property
    def task(self):
        if self.type == 'row':
            return self._build_elementwise_function()
        return self._build_vectorised_function()

    @cached_property
    def fn(self) -> Callable:
        if len(self.funcs) == 1:
            return self.funcs[0]

        def compose(f: Callable[[pd.Series], pd.Series],
                    g: Callable[[pd.Series], pd.Series]):
            return lambda x: g(f(x))

        return reduce(compose, self.funcs, lambda x: x)

    @cached_property
    def name(self) -> str:
        return ','.join(self.features)
